"""Definitions for specifying how to start a component."""
from typing import cast, List, Union

from ruamel import yaml
import yatiml

from ymmsl.identity import Reference


class Implementation:
    """Describes an installed implementation.

    Attributes:
        name: Name of the implementation
        script (str): A script that starts the implementation.
    """

    def __init__(
            self, name: Reference, script: Union[str, List[str]]) -> None:
        """Create an Implementation description.

        If script is a list, each string in it is a line, and the
        lines will be concatenated into a single string to put into
        the script attribute.

        Args:
            name: Name of the implementation
            script: Script that starts the implementation
        """
        self.name = name

        if isinstance(script, list):
            self.script = '\n'.join(script)
        else:
            self.script = script

    @classmethod
    def _yatiml_sweeten(cls, node: yatiml.Node) -> None:
        script_node = node.get_attribute('script')
        if script_node.is_scalar(str):
            text = cast(str, script_node.get_value())
            if '\n' in text:
                # make a sequence of lines
                lines = text.split('\n')
                start_mark = node.yaml_node.start_mark
                end_mark = node.yaml_node.end_mark
                lines_nodes = [
                        yaml.ScalarNode(
                            'tag:yaml.org,2002:str', line, start_mark,
                            end_mark)
                        for line in lines]
                seq_node = yaml.SequenceNode(
                        'tag:yaml.org,2002:seq', lines_nodes, start_mark,
                        end_mark)
                node.set_attribute('script', seq_node)


class Resources:
    """Describes resources to allocate for components.

    Attributes:
        name: Name of the component to configure.
        num_cores: Number of CPU cores to reserve.
    """

    def __init__(self, name: Reference, num_cores: int) -> None:
        """Create a Resources description.

        Args:
            name: Name of the component to configure
            num_cores: Number of CPU cores to reserve.
        """
        self.name = name
        self.num_cores = num_cores



"""Loading and saving functions."""
from pathlib import Path
from typing import Any, IO, Union

import yatiml

from ymmsl.configuration import Configuration, PartialConfiguration
from ymmsl.document import Document
from ymmsl.execution import Implementation, Resources
from ymmsl.settings import Settings
from ymmsl.identity import Identifier, Reference
from ymmsl.model import Component, Conduit, Model, ModelReference


_load = yatiml.load_function(
        PartialConfiguration, Component, Conduit, Configuration, Document,
        Identifier, Implementation, Model, ModelReference, Reference,
        Resources, Settings)


def load(source: Union[str, Path, IO[Any]]) -> PartialConfiguration:
    """Loads a yMMSL document from a string or a file.

    Args:
        source: A string containing yMMSL data, a pathlib Path to a
                file containing yMMSL data, or an open file-like
                object containing from which yMMSL data can be read.

    Returns:
        A PartialConfiguration object corresponding to the input data.

    """
    # This wrapper is just here to render the documentation.
    return _load(source)


_dump = yatiml.dumps_function(
        Component, Conduit, Configuration, Document, Identifier,
        Implementation, Model, ModelReference, PartialConfiguration, Reference,
        Resources, Settings)


def dump(config: PartialConfiguration) -> str:
    """Converts a Ymmsl to a string containing YAML.

    Args:
        config: The configuration to be saved to yMMSL.

    Returns:
        A yMMSL YAML description of the given document.

    """
    # This wrapper is just here to render the documentation.
    return _dump(config)


_save = yatiml.dump_function(
        Component, Conduit, Configuration, Document, Identifier,
        Implementation, Model, ModelReference, PartialConfiguration, Reference,
        Resources, Settings)


def save(
        config: PartialConfiguration, target: Union[str, Path, IO[Any]]
        ) -> None:
    """Saves a yMMSL document to a file.

    Args:
        config: The configuration to save to yMMSL.
        target: The file to save to, either as a string containing a
            path, as a pathlib Path object, or as an open file-like
            object.

    """
    # This wrapper is just here to render the documentation.
    _save(config, target)



"""Definitions for describing simulation components."""
from enum import Enum

from ymmsl.identity import Identifier


class Operator(Enum):
    """An operator of a compute element.

    This is a combination of the Submodel Execution Loop operators,
    and operators for other components such as mappers.
    """

    NONE = 0    #: No operator
    F_INIT = 1  #: Initialisation phase, before start of the SEL
    O_I = 2     #: State observation within the model's main loop
    S = 3       #: State update in the model's main loop
    B = 4       #: Boundary conditions update in the model's main loop
    O_F = 5     #: Observation of final state, after the SEL

    def allows_sending(self) -> bool:
        """Whether ports on this operator can send."""
        return self in {
                Operator.NONE,
                Operator.O_I,
                Operator.O_F
        }

    def allows_receiving(self) -> bool:
        """Whether ports on this operator can receive."""
        return self in {
                Operator.NONE,
                Operator.F_INIT,
                Operator.S,
                Operator.B
                }


class Port:
    """A port on a compute element.

    Ports are used by compute elements to send or receive messages
    on. They are connected by conduits to enable communication between
    compute elements.

    Attributes:
        name: The name of the port.
        operator: The MMSL operator in which this port is used.

    """

    def __init__(self, name: Identifier, operator: Operator) -> None:
        """Create a Port.

        Args:
            name: The name of the port.
            operator: The MMSL Operator in which this port is used.

        """
        self.name = name  # type: Identifier
        self.operator = operator  # type: Operator



"""Python bindings for yMMSL.

This package contains all the classes needed to represent a yMMSL file,
as well as to read and write yMMSL files.
"""

from ymmsl.component import Operator, Port
from ymmsl.configuration import Configuration, PartialConfiguration
from ymmsl.execution import Implementation, Resources
from ymmsl.settings import Settings, SettingValue
from ymmsl.identity import Identifier, Reference
from ymmsl.io import load, dump, save
from ymmsl.model import Component, Conduit, Model, ModelReference


__version__ = '0.11.0'
__author__ = 'Lourens Veen'
__email__ = 'l.veen@esciencecenter.nl'


__all__ = [
        'Component', 'Conduit', 'Configuration', 'dump', 'Identifier',
        'Implementation', 'load', 'Model', 'ModelReference', 'Operator',
        'PartialConfiguration', 'Resources', 'SettingValue', 'Port',
        'Reference', 'Settings', 'save']
